/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS AS-IS
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */

/* based on the mips/PMAX/kmin_dma.c code */

#include <mach/std_types.h>
#include <sys/types.h>
#include <chips/busses.h>
#include <machine/machparam.h>		/* spl definitions */
#include <device/io_req.h>
#include <device/conf.h>

#include <machine/a3000.h>
#include <machine/scsi.h>

#define NA3000	1	/* XXX */

/*
 * Definition of the driver for the auto-configuration program.
 */

int	a3000_probe(), a3000_attach(), a3000_intr();


/* a bit ridiculous I know... */
caddr_t	a3000_std[NA3000] = { 0 };
struct	bus_device *a3000_info[NA3000];
struct	bus_driver a3000_driver = 
        { a3000_probe, 0, a3000_attach, 0, a3000_std, "a3000", a3000_info,};

/*
 * Adapt/Probe/Attach functions
 */
a3000_probe (xxx, ui)
    struct bus_device *ui;
{
  /* we are only called if ??? hand-filled us in */
  return 1;
}

a3000_attach(ui)
	register struct bus_device *ui;
{
  /* we got nothing to say */
}

a3000_intr (unit, spllevel, intr)
{
  if (a3000_dmac.int_status & A3000_DMAC_INT_PENDING)
    a3000_dma_intr (unit, spllevel, intr);
}


/*
 *	Functions to support DMA for the on-board SCSI
 */

#include <scsi/compat_30.h>
#include <scsi/scsi.h>
#include <scsi/scsi_defs.h>
#include <scsi/adapters/scsi_dma.h>
#include <scsi/adapters/scsi_33C93.h>

/* utilities */

#define	u_min(a,b)	(((a) < (b)) ? (a) : (b))

int dma_in_progress = 0;

static inline void 
a3000_stop_dma ()
{
  if (dma_in_progress)
    {
      a3000_dmac.flush_DMA = 1;
      while (! (a3000_dmac.int_status & A3000_DMAC_DMA_IDLE)) ;
      a3000_dmac.clear_ints = 1;
      a3000_dmac.stop_DMA = 1;
      
      dma_in_progress = 0;
    }
}  

static inline void *
kvtophys (vm_offset_t kv)
{
  extern pmap_t kernel_pmap;
  
  return pmap_extract (kernel_pmap, kv);
}

a3000_dma_read (dma_ptr, count)
    vm_offset_t	dma_ptr;
{
  register vm_offset_t	phys;

  /* stop DMA engine first */
  a3000_stop_dma ();

  phys = kvtophys (dma_ptr);

  a3000_dmac.control = A3000_DMAC_DIR_READ;
  a3000_dmac.DMA_address = phys;
  a3000_dmac.start_DMA = 1;
  
  dma_in_progress = 1;
}

a3000_dma_write (dma_ptr, count)
    vm_offset_t	dma_ptr;
{
  register vm_offset_t	phys;

  /* stop DMA engine first */
  a3000_stop_dma ();

  phys = kvtophys(dma_ptr);

  a3000_dmac.control = A3000_DMAC_DIR_WRITE;
  a3000_dmac.DMA_address = phys;
  a3000_dmac.start_DMA = 1;
  
  dma_in_progress = 1;
}

a3000_dma_intr (unit, spllevel, intr)
{
  a3000_stop_dma ();
}


/*
 * Cold, probe time init
 */
opaque_t
a3000_dma_init (unit, base, dma_bsizep)
    int		unit;
    vm_offset_t	base;
    int		*dma_bsizep;
{
  a3000_dmac.DACK_width = A3000_DMAC_WIDTH;
  return (opaque_t) 0;
}

/*
 * A target exists
 */
void
a3000_dma_new_target (dma, tgt)
    opaque_t	dma;
    target_info_t	*tgt;
{
  /* provide enough space here, the cmd_buf is used for both, commands and
     (small) results of maintanance commands (ie. inquiry)
     MAKE SURE THE BUFFERS ARE LONG ALIGNED!! */
  static u_int cmds[NSCSI][8][256/4];

  tgt->cmd_ptr = (u_char *) cmds[tgt->masterno][tgt->target_id];
}

/*
 * Map virtual addresses onto physical ones
 */
void
a3000_dma_map (dma, tgt)
    opaque_t	dma;
    target_info_t	*tgt;
{
  /*
   * We do real DMA.
   */

  if ((tgt->cur_cmd == SCSI_CMD_WRITE) ||
      (tgt->cur_cmd == SCSI_CMD_LONG_WRITE))
    {
      io_req_t	   ior = tgt->ior;
      register int len = ior->io_count;

      tgt->transient_state.out_count = len;

      /* avoid leaks ?? */
      if (len < tgt->block_size) 
	{
/*	  bzero(to + len, tgt->block_size - len);*/
	  len = tgt->block_size;
	  tgt->transient_state.out_count = len;
	}

      tgt->transient_state.dma_offset = 0;
    } 
  else 
    {
      tgt->transient_state.out_count = 0;
      tgt->transient_state.dma_offset = 0;
    }
}

int
a3000_dma_start_cmd (dma, tgt)
    opaque_t	         dma;
    register target_info_t *tgt;
{
  /* contrary to the 3min, this command already starts DMA for the data
     transfer, since there's no CDB to be transmitted */

  if (tgt->transient_state.in_count)
    return a3000_dma_start_datain (dma, tgt);
  else if (tgt->transient_state.out_count)
    return a3000_dma_start_dataout (dma, tgt);
  else
    return 0;
}

void
a3000_dma_end_cmd(dma, tgt, ior)
    opaque_t	  dma;
    target_info_t *tgt;
    io_req_t	  ior;
{
  tgt->dma_ptr = 0;
  tgt->transient_state.dma_offset = 0;
}

void
a3000_dma_end_xfer (dma, tgt, bytes_read)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  bytes_read;
{
  register volatile char  *ptr;
  register int		  nb;

  a3000_stop_dma ();

  /* On read, flush cache (necessary with our DMAC??) XXXX */
  if (bytes_read != 0)
    {
      /* would be nice to do this in parallel with xfer, earlier on */
      ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;
      flush_the_damn_cache (ptr, bytes_read);
    }
}

int
a3000_dma_start_datain (dma, tgt)
    opaque_t		   dma;
    register target_info_t *tgt;
{
  register char	*dma_ptr;
  register int	count;

  /* setup tgt->dma_ptr to ior->io_data, flush cache  */
  if (tgt->cur_cmd == SCSI_CMD_READ || tgt->cur_cmd == SCSI_CMD_LONG_READ)
    dma_ptr = tgt->ior->io_data;
  else
    /* swear it won't disconnect midway a xfer or else */
    dma_ptr = tgt->cmd_ptr;

  if (!(tgt->dma_ptr = dma_ptr))
    panic ("a3000_dma_start_datain, no dma_ptr");

  count = tgt->transient_state.in_count;
  if (count <= 0)
    panic ("a3000_dma_start_datain");

  a3000_dma_read(dma_ptr, count);

#if 0
  count = u_min(count, SBIC_TC_MAX);
#else
  /* too bad the A3000 DMA can't chain DMA requests ;-( */
  count = u_min (count, PMAP_PGBYTES);
#endif

  return count;
}

int
a3000_dma_restart_datain_1 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  register char *dma_ptr;
  register int  count;

  /* This is the unlikely case we disconnected before
     actually starting the data xfer */
  if (tgt->dma_ptr == 0)
    return a3000_dma_start_datain (dma, tgt);

  /* keep dma_offset anyways */
  dma_ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;

  count = tgt->transient_state.in_count;

  a3000_dma_read (dma_ptr, count);

#if 0
  count = u_min(count, SBIC_TC_MAX);
#else
  /* too bad the A3000 DMA can't chain DMA requests ;-( */
  count = u_min (count, PMAP_PGBYTES);
#endif
  return count;
}

int
a3000_dma_restart_datain_2 (dma, tgt, xferred)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  xferred;
{
/*XXX*/	tgt->transient_state.copy_count = tgt->transient_state.dma_offset;
  tgt->transient_state.dma_offset += xferred;

  return a3000_dma_restart_datain_1 (dma, tgt);
}

void
a3000_dma_restart_datain_3 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  register volatile char  *ptr;
  int			  len;

  /* avoid doing it again!! XXXX */
  ptr = tgt->dma_ptr + tgt->transient_state.copy_count;
  len = tgt->transient_state.dma_offset + tgt->transient_state.copy_count;
  flush_the_damn_cache (ptr, len);
}


boolean_t
a3000_dma_start_dataout (dma, tgt, regp, cmd)
    opaque_t		   dma;
    register target_info_t *tgt;
    volatile char	   *regp, cmd;
{
  register char	*dma_ptr;

  /* since we don't transmit the CDB explicitly, make sure no code depends
     on such dma stuff */
  if (tgt->transient_state.dma_offset)
    panic ("a3000_dma_start_dataout");

  dma_ptr = tgt->ior->io_data;
  if (! dma_ptr)
    panic ("a3000_dma_start_dataout no data_ptr!");

  a3000_dma_write (dma_ptr, tgt->transient_state.out_count);

  tgt->transient_state.dma_offset = 0;
  tgt->dma_ptr = dma_ptr;

  return TRUE;
}

int
a3000_dma_restart_dataout_1 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  int	count;

  /* this is if we disconnected before starting any xfer */
  if (tgt->dma_ptr == 0)
    a3000_dma_start_dataout (dma, tgt, &count, 0);

  count = tgt->transient_state.out_count;
#if 0
  count = u_min(count, SBIC_TC_MAX);
#else
  /* too bad the A3000 DMA can't chain DMA requests ;-( */
  count = u_min (count, PMAP_PGBYTES);
#endif

  return count;
}

int
a3000_dma_restart_dataout_2 (dma, tgt, xferred)
    opaque_t		   dma;
    register target_info_t *tgt;
    int			   xferred;
{
  tgt->transient_state.dma_offset += xferred;

  return a3000_dma_restart_dataout_1 (dma, tgt);
}

int
a3000_dma_restart_dataout_3 (dma, tgt, regp)
    opaque_t	  dma;
    target_info_t *tgt;
    volatile char *regp;
{
  register char	*dma_ptr;
  int		ret;

  /* keep dma_offset anyways */
  dma_ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;

/* unaligned issues here, align to 8 bytes ? (maybe not..) XXXX */

  a3000_dma_write (dma_ptr, tgt->transient_state.out_count);

  return 0; /* or 1 if adjusted */
}

void
a3000_dma_restart_dataout_4 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* Nothing needed here */
}

int
a3000_dma_start_msgin (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* SBIC ONLY SUPPORTS DATA DMA, NO MSGS */
  return 0;
}

void
a3000_dma_end_msgin (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* SBIC ONLY SUPPORTS DATA DMA, NO MSGS */
}


boolean_t
a3000_dma_disconn_1 (dma, tgt, xferred)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  xferred;
{
  /*
   * XXX Known bug: if we disconnect on an odd bb ??
   */
  a3000_stop_dma ();

  tgt->transient_state.dma_offset += xferred;
  return FALSE;
}

boolean_t
a3000_dma_disconn_2 (dma, tgt)
    opaque_t	  dma;
    target_info_t *tgt;
{
  /* Nothing needed */
  return FALSE;
}


boolean_t
a3000_dma_disconn_3 (dma, tgt, xferred)
    opaque_t	  dma;
    target_info_t *tgt;
    int		  xferred;
{
  register volatile char	*ptr;

  a3000_stop_dma ();

  ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;
  flush_the_damn_cache (ptr, xferred);

  tgt->transient_state.dma_offset += xferred;
  return FALSE;
}

#define	a3000_dma_disconn_4		a3000_dma_disconn_3
#define	a3000_dma_disconn_5		a3000_dma_disconn_1
#define	a3000_dma_disconn_callback	0	/* never called */

flush_the_damn_cache (p, nb)
    vm_offset_t p;
{
  vm_size_t off;

  off = p & (PMAP_PGBYTES-1);
  off = PMAP_PGBYTES - off;
  while (nb) 
    {
      off = (nb < off) ? nb : off;
#if 0
      mipscache_Dflush( PHYS_TO_K0SEG(kvtophys(p)), off);
#else
      /* XXXXX HOW DO YOU DO THIS WITH M68K PMAP CODE??? 
         moveq #??,d0
         movec d0,cacr
         ??????? */
#endif
      p += off;
      nb -= off;
      off = PMAP_PGBYTES;
    }
}

scsi_dma_ops_t a3000_dma_ops = {
  a3000_dma_init,
  a3000_dma_new_target,
  a3000_dma_map,
  a3000_dma_start_cmd,
  a3000_dma_end_xfer,
  a3000_dma_end_cmd,
  a3000_dma_start_datain,
  a3000_dma_start_msgin,
  a3000_dma_end_msgin,
  a3000_dma_start_dataout,
  a3000_dma_restart_datain_1,
  a3000_dma_restart_datain_2,
  a3000_dma_restart_datain_3,
  a3000_dma_restart_dataout_1,
  a3000_dma_restart_dataout_2,
  a3000_dma_restart_dataout_3,
  a3000_dma_restart_dataout_4,
  a3000_dma_disconn_1,
  a3000_dma_disconn_2,
  a3000_dma_disconn_3,
  a3000_dma_disconn_4,
  a3000_dma_disconn_5,
  a3000_dma_disconn_callback,
};

/* moved here so I don't have to declare a3000_dma_ops in the header... */
a3000_init ()
{
  a3000_dmac.DACK_width = A3000_DMAC_WIDTH;
  sbic_set_dmaops (0, &a3000_dma_ops);
}

